///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Filename:		Main.asm
//							(c) ANALOG DEVICES 2001, gjf
//
// Date:				10/23/2001
// Purpose:			Software Overlay Main Program Assembler Example
//
// Inputs:			none
// Outputs:			none
//
// Description:	This progam calls three simple assembly routines each of which reside in external 8-bit SRAM memory.
//							The calls to these assembly functions invoke the soft overlay manager which manages the loading of
//							these overlays from external memory into internal memory via the DMA engine of the 2191. The overlay
//							manager also is responsible for making the program sequencer jump to the loaded overlay region's
//							target address upon completion of the DMA. For this example, ALL of the overlay regions 'live' in
//							external memory and 'run' from internal memory space.
//
//							This program also calls the overlay manager pre-emptively to load in code and data overlay modules
//							from external memory before they are needed or executed to increase system performance and to
//							minimize latency.
//
// Registers Used:
//
//							primary register set:		ax0, sr2, sr1, sr0, ar, dmpg2, iopg, IJPG (not saved or restored!!!!)
//							primary DAGs:						i7, m7
//							secondary register set:	none
//							secondary DAGs:					none
//
// Additional Information:
//
//							none
//
// Revision Information:
//
//							rev 0.0 - (10/23/01) -	This revision assumes that ALL calls to the loaded overlay modules are within internal
//																			PM memory. Therefore the IJPG register is NEVER saved or restored. If the user wishes
//																			to execute PM overlays that 'run' from external memory, then care must be taken to
//																			ensure that the IJPG register is properly saved and restored on the overlay stack.
//
//											- (10/31/01) -	The call from main (this program) for the overlay module that we are to fetch calls the
//																			desired overlay by it's actual ID# + 1. For example, if we wish to fetch the 2nd PM
//																			overlay module, we would load "2" into ax0 before calling the overlay manager even though
//																			the actual overlay ID# for the 2nd PM overlay module is actually 1. ("Off by one" value,
//																			just like when using arrays in C.)
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// The definitions below used for readability when fetching an overlay region

#define Function_1_Add		0x0001
#define Function_2_Mult		0x0002
#define Function_3_Sub		0x0003
#define Function_4_Junk		0x0004

#define DM_Overlay_1			0x0005
#define DM_Overlay_2			0x0006
#define DM_Overlay_3			0x0007

#define	Fetch_Overlay(x) \
				dm(save_ax0) = ax0;\
				dm(save_ax1) = ax1;\
				dm(save_ay1) = ay1;\
				ax0 = x;\
				dm(ov_stack) = ax0;\
				call _OverlayManager;\
				ax0 = dm(save_ax0);\
				ax1 = dm(save_ax1);\
				ay1 = dm(save_ay1);


#include "2191asm.h"										// include file containing register and bit definitions

.extern func1_add;											// function which 'lives' in external memory and 'runs' in internal PM
.extern func2_mult;											// function which 'lives' in external memory and 'runs' in internal PM
.extern func3_sub;											// function which 'lives' in external memory and 'runs' in internal PM
.extern _OverlayManager;								// Address label of the beginning of the overlay manager
.extern save_ax0;												// temporary memory locations for saving-restoring of ax0 register
.extern save_ay0;												// temporary memory locations for saving-restoring of ay0 register
.extern ovly_ID_number;									// variable used to check if the overlay region is already loaded into internal run memory space
.extern ovly_ID_temp;
.extern ov_stack;												// this is a software stack used to save & restore registers clobbered by the overlay manager
.extern curr_PM_ovly_ID1;								// variable used to check if the PM overlay region is already loaded into internal PM run memory space #1
.extern curr_PM_ovly_ID2;								// variable used to check if the PM overlay region is already loaded into internal PM run memory space #2
.extern curr_DM_ovly_ID1;								// variable used to check if the DM overlay region is already loaded into internal DM run memory space #1
.extern curr_DM_ovly_ID2;								// variable used to check if the DM overlay region is already loaded into internal DM run memory space #2
.extern LRU_PM_Ovly_1;									// LRU variable used to define if internal PM run memory space #1 is least recently used or not
.extern LRU_PM_Ovly_2;									// LRU variable used to define if internal PM run memory space #2 is least recently used or not
.extern LRU_DM_Ovly_1;									// LRU variable used to define if internal DM run memory space #1 is least recently used or not
.extern LRU_DM_Ovly_2;									// LRU variable used to define if internal DM run memory space #2 is least recently used or not
.extern save_ax1;												// temporary memory locations for saving-restoring of ax1 register
.extern save_ay1;												// temporary memory locations for saving-restoring of ay1 register


.section/pm	seg_code;
.global start;
start:

	nop;																	// NOP added here for debugging purposes only
	
Clear_Regs:

	nop;																	// NOP added here for debugging purposes only

#ifdef DEBUG
	
	// just clear all of the registers for debugging purposes and for a clean environment !!!!!
	ar = 0x0000;
	
	// clear ALU registers
	ax0 = ar; ay0 = ar; ax1 = ar; ay1 = ar;
	
	// clear DAG registers
	i0 = ar; i1 = ar; i2 = ar; i3 = ar; i4 = ar; i5 = ar; i6 = ar; i7 = ar;
	m0 = ar; m1 = ar; m2 = ar; m3 = ar; m4 = ar; m5 = ar; m6 = ar; m7 = ar;
	l0 = ar; l1 = ar; l2 = ar; l3 = ar; l4 = ar; l5 = ar; l6 = ar; l7 = ar;
	reg(b0) = ar; reg(b1) = ar; reg(b2) = ar; reg(b3) = ar; reg(b4) = ar; reg(b5) = ar; reg(b6) = ar; reg(b7) = ar; 
	
	// clear MAC registers
	mx0 = ar; mx1 = ar; my0 = ar; my1 = ar;
	mr0 = ar; mr1 = ar; mr2 = ar;
	
	// clear shifter registers
	si = ar; sb = ar; se = ar;
	sr0 = ar; sr1 = ar; sr2 = ar;
	
#endif DEBUG

Interrupt_Priority_Config:

	/////////////////////////////////////////////////////////////////////////////////////////////////
	// Assign all of the peripheral interrupt priorities (EXCEPT for MEM DMA) to the lowest
	// priority setting. MEM DMA will be assigned the highest interrupt priority level. This is
	// strictly for academic purposes since in this example, the only active interrupt will be
	// MEM DMA.
	/////////////////////////////////////////////////////////////////////////////////////////////////

	iopg = Clock_and_System_Control_Page;				// switch to iopg 0 for system configuration register
	ar = io(SYSCR);															// read value of system configuration register into ar
	ar = setbit 4 of ar;												// execute from on-chip SRAM starting at address 0x000000 (PM)

	#ifdef BUS_WIDTH_8
	ar = setbit 5 of ar;												// enable PF15:8 (8-bit external memory bus)
	#endif BUS_WIDTH_8
	
	io(SYSCR) = ar;															// write these values back into the system configuration register

	dis int;																		// globally disable all interrupts (we'll enable interrupts in the overlay manager for MEM DMA)
	imask = 0;																	// mask all interrupts

	iopg = Interrupt_Controller_Page;						// switch to the interrupt controller page (0x01)
	ax0 = 0xbbbb;																// set ax0 to lowest priority interrupt level
	io(IPR0) = ax0;															// assign lowest priority to host, SPORT0 rx/tx, & SPORT1 rx interrupts
	io(IPR1) = ax0;															// assign lowest priority to SPORT1 tx, SPORT2 rx/tx, & UART rx interrupts
	io(IPR2) = ax0;															// assign lowest priority to UART tx, timer 0-2 interrupts
	ax0 = 0xb0bb;
	io(IPR3) = ax0;															// assign highest priority to MEM DMA, lowest priority to FLAG A & B interrupts
	
	icntl = 0;																	// disable interrupt nesting
	irptl = 0;																	// clear all (pending) interrupts
	imask = 0x0010;															// allow only priority #4 interrupt (we've assigner #4 to MEM DMA)

	nop;																				// NOP added here for debug purposes only

Configure_EMI:

	/////////////////////////////////////////////////////////////////////////////////////////////////
	// Here we configure the EMI to model the bus width of the external memory that we have in our
	// system where the overlay modules reside. Currently there is a hardware limitation that
	// restricts us to only supporting the initialization of external memory which is the same
	// physical width as the memory device that we boot from. For our example, since we are booting
	// from a byte-wide EPROM, we MUST also initialize the overlay modules to reside in byte-wide
	// SRAM. (gjf 10-29-01)
	//
	// The conditional macro "BUS_WIDTH_8" is defined below to facilitate easy code modification to
	// change the bus width of the external SRAM to either 8- or 16-bits.
	/////////////////////////////////////////////////////////////////////////////////////////////////

	iopg = External_Memory_Interface_Page;			// switch to the EMI interface page (0x06)
	
	#ifdef BUS_WIDTH_8
	ax0 = 0x0030;																// configure EMI for 8-bit ext bus, ~RD/~WR acive low, DMA split disabled
	#else
	ax0 = 0x0038;																// configure EMI for 16-bit ext bus, ~RD/~WR acive low, DMA split disabled
	#endif BUS_WIDTH_8
	
	io(EMICTL) = ax0;

	/////////////////////////////////////////////////////////////////////////////////////////////////
	// Assign all of the mem space ctrl regs to zero wait states, internal wait mode only,
	// 1x core clock base clock div, and no wait hold cycle
	/////////////////////////////////////////////////////////////////////////////////////////////////

	ax0 = 0x0040;
	io(MS0CTL) = ax0;														// Memory Space Bank 0 Access Control Register
	io(MS1CTL) = ax0;														// Memory Space Bank 1 Access Control Register
	io(MS2CTL) = ax0;														// Memory Space Bank 2 Access Control Register
	io(MS3CTL) = ax0;														// Memory Space Bank 3 Access Control Register
	io(IOMSCTL) = ax0;													// IO Space Access Control Register
	
	iopg = External_Access_Bridge_Page;					// set to IOPG 0x00
	ax0 = 0x0008;
	io(E_STAT) = ax0;														// switch to 24-bit data accesses =(0x0008): 16-bit = 0x0000


	/////////////////////////////////////////////////////////////////////////////////////////////////
	// Begin main program code below...
	/////////////////////////////////////////////////////////////////////////////////////////////////

Start_Program:

	ax0 = 0xdead;														// assign some arbitrary value to ax0 (for debugging purposes only)
	ax1 = 0x0005;														// assign some arbitrary value to ax1
	ay1 = 0x0003;														// assign some arbitrary value to ay1
	
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Check here to see if the DM overlay that we need is loaded into internal memory or not. Since this is the first
	// overlay to be loaded in, we know that it HAS to be loaded in by the overlay manager, but this code is just
	// placed here for illustration of this procedure. The variables "curr_DM_ovly_ID1" and "curr_DM_ovly_ID2" should
	// be zero by default (NULL) at the beginning of program execution since no DM overlay has been loaded yet.
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	

Fetch_DM_Ovly_3:

	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here we need to jump to the overlay manager, but first we need to preload the value of ax0 (the overlay ID#) into the 
	// overlay manager in order to first determine whether the overlay already exists within an overlay "run" segment already,
	// and if the overlay module does not already exist in one of the overlay "run" segment then fetch it.
	/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	Fetch_Overlay(DM_Overlay_3);						// we know that we want to fetch the third DM overlay module (#6 - 3 PM ovlys, 3 DM ovlys)
	
	nop;																		// NOP added here for no real reason...

	/////////////////////////////////////////////////////////////////////////////////////////
	// At this point we can use register ax0 for our main program. (Place your code here...)
	/////////////////////////////////////////////////////////////////////////////////////////
	
	nop;																		// NOPs added here for no particular reason whatsoever...
	nop;
	nop;
	
Goto_Function_1:

	////////////////////////////////////////////////////////////////////////////////////////////////////
	// Here we need to call the overlay manager again to load in our first function (Add). Again, since
	// this is the first PM overlay function that is to be loaded into one of the internal PM overlay
	// run spaces, we will need to load in this overlay module again before we can begin executing it.
	////////////////////////////////////////////////////////////////////////////////////////////////////

	Fetch_Overlay(Function_1_Add);				// Call the function Add (since this is the first call to this function we must load it in via the overlay manager)

	nop;																	// NOP added here for no particular reason

Goto_Function_1_Again:

	Fetch_Overlay(Function_1_Add);				// call function #1 again (it should already be in memory at this point in the main program)

Fetch_DM_Ovly_3_Again:

	Fetch_Overlay(DM_Overlay_3);
	nop;

Fetch_DM_Ovly_1:

	Fetch_Overlay(DM_Overlay_1);					// fetch the data overlay #1 (first call to this data overlay)

Goto_Function_2:

	Fetch_Overlay(Function_2_Mult);				// call overlay function #2 (multiply function)

	nop;																	// NOP added here for no particular reason

Fetch_DM_Ovly_2:

	Fetch_Overlay(DM_Overlay_2);					// 

Goto_Function_3:

	Fetch_Overlay(Function_3_Sub);				// call overlay function #3 (subtract function)
	nop;

Goto_Function_4:

	Fetch_Overlay(Function_4_Junk);				// call overlay function #4
	nop;
	
Goto_Function_3_Again:

	Fetch_Overlay(Function_3_Sub);				// call overlay function #3 again (subtract function)

Goto_Function_4_Again:

	Fetch_Overlay(Function_4_Junk);				// call overlay function #4 again
	nop;

	nop;																	// NOPs added here for debug only
	nop;
	nop;
	nop;
	nop;
					
end:	

	nop;																	// NOP added here for no particular reason
	nop;																	// NOP added here for no particular reason
	nop;																	// NOP added here for no particular reason
	nop;																	// NOP added here for no particular reason
	nop;																	// NOP added here for no particular reason
	
	jump end;
